import contextlib

try:
    import importlib.resources as resources
except ImportError:  # for Python<3.7
    import importlib_resources as resources


class IconDist:

    def __init__(self):
        self._context = contextlib.ExitStack()
        self._paths = {}

    def patch_list(self, icons):
        return [
            path
            for icon in icons
            for path in [icon, self.lookup(icon)]
            if path
        ]

    def lookup(self, name):
        if name in self._paths:
            return self._paths[name]
        try:
            path = self._context.enter_context(
                resources.path('udiskie.icons', name + '.svg'))
            path = str(path.absolute())
        except FileNotFoundError:
            path = None
        self._paths[name] = path
        return path



"""
Tests for the udiskie.keyutils wrapper.
"""

import unittest
import time

import udiskie.keyutils as keyutils


class TestKeyutils(unittest.TestCase):

    """
    Tests for the udiskie.keyutils wrapper.
    """

    # NOTE: The key names are different in each test so that they do not
    # interfere accidentally.

    def setUp(self):
        keyutils.clear()

    def test_add_request_read(self):
        """The cached password expires after the specified timeout."""
        key = b'ALPHA'
        val = '{<}hëllo ωορλδ!{>}'.encode('utf-8')
        key_id = keyutils.add_key(key, val)
        self.assertEqual(keyutils.request_key(key), key_id)
        self.assertEqual(keyutils.read_key(key_id), val)

    def test_touch(self):
        """Key access refreshes the timeout."""
        key = b'BETA'
        val = '{<}hëllo ωορλδ!{>}'.encode('utf-8')
        key_id = keyutils.add_key(key, val)
        self.assertEqual(keyutils.request_key(key), key_id)
        self.assertEqual(keyutils.read_key(key_id), val)

        keyutils.set_timeout(key_id, 3)
        time.sleep(1)
        self.assertEqual(keyutils.read_key(key_id), val)

        keyutils.set_timeout(key_id, 3)
        time.sleep(2)
        self.assertEqual(keyutils.read_key(key_id), val)

        keyutils.set_timeout(key_id, 3)
        time.sleep(4)
        with self.assertRaises(keyutils.KeyExpired):
            keyutils.read_key(key_id)

    def test_invalidate(self):
        """A key can be deleted manually."""
        key = b'GAMMA'
        val = '{<}hëllo ωορλδ!{>}'.encode('utf-8')
        key_id = keyutils.add_key(key, val)
        self.assertEqual(keyutils.request_key(key), key_id)
        self.assertEqual(keyutils.read_key(key_id), val)

        keyutils.invalidate(key_id)
        with self.assertRaises(keyutils.KeyutilsError):
            keyutils.read_key(key_id)
        with self.assertRaises(keyutils.KeyNotAvailable):
            keyutils.request_key(key)

    def test_update(self):
        key = b'DELTA'
        val = '{<}hëllo ωορλδ!{>}'.encode('utf-8')
        key_id = keyutils.add_key(key, val)
        self.assertEqual(keyutils.request_key(key), key_id)
        self.assertEqual(keyutils.read_key(key_id), val)

        key_id = keyutils.add_key(key, val + val)
        self.assertEqual(keyutils.request_key(key), key_id)
        self.assertEqual(keyutils.read_key(key_id), val + val)

    def test_clear(self):
        key1 = b'eps'
        key2 = b'tau'
        val1 = b'hello'
        val2 = b'world'
        id1 = keyutils.add_key(key1, val1)
        id2 = keyutils.add_key(key2, val2)
        self.assertEqual(keyutils.request_key(key1), id1)
        self.assertEqual(keyutils.request_key(key2), id2)
        self.assertEqual(keyutils.read_key(id1), val1)
        self.assertEqual(keyutils.read_key(id2), val2)

        keyutils.clear()
        with self.assertRaises(keyutils.KeyutilsError):
            keyutils.read_key(id1)
        with self.assertRaises(keyutils.KeyutilsError):
            keyutils.read_key(id2)

    def test_list_keys(self):
        self.assertCountEqual([], keyutils.list_keys())

        id1 = keyutils.add_key(b"rho1", b"val1")
        id2 = keyutils.add_key(b"rho2", b"val2")
        self.assertCountEqual([id1, id2], keyutils.list_keys())

    def test_is_valid(self):
        unknown_key = 123456
        self.assertFalse(keyutils.is_valid(unknown_key))

        key_id = keyutils.add_key(b"gamma", b"value")
        self.assertTrue(keyutils.is_valid(key_id))

        keyutils.invalidate(key_id)
        self.assertFalse(keyutils.is_valid(key_id))

        key_id = keyutils.add_key(b"gamma", b"value")
        self.assertTrue(keyutils.is_valid(key_id))

        keyutils.set_timeout(key_id, 1)
        time.sleep(1.2)
        self.assertFalse(keyutils.is_valid(key_id))


if __name__ == '__main__':
    unittest.main()



"""
Automount utility.
"""

from .common import DaemonBase
from .async_ import run_bg


__all__ = ['AutoMounter']


class AutoMounter(DaemonBase):

    """
    Automount utility.

    Being connected to the udiskie daemon, this component automatically
    mounts newly discovered external devices. Instances are constructed with
    a Mounter object, like so:

    >>> automounter = AutoMounter(Mounter(udisks=Daemon()))
    >>> automounter.activate()
    """

    def __init__(self, mounter, automount=True):
        """Store mounter as member variable."""
        self._mounter = mounter
        self._automount = automount
        self.events = {
            'device_changed': self.device_changed,
            'device_added': self.auto_add,
            'media_added': self.auto_add,
        }

    def is_on(self):
        return self._automount

    def toggle_on(self):
        self._automount = not self._automount

    def device_changed(self, old_state, new_state):
        """Mount newly mountable devices."""
        # udisks2 sometimes adds empty devices and later updates them - which
        # makes is_external become true at a time later than device_added:
        if (self._mounter.is_addable(new_state)
                and not self._mounter.is_addable(old_state)
                and not self._mounter.is_removable(old_state)):
            self.auto_add(new_state)

    @run_bg
    def auto_add(self, device):
        return self._mounter.auto_add(device, automount=self._automount)



"""
ctypes wrapper for keyutils functionality used in udiskie.
"""

__all__ = [
    "add_key",
    "read_key",
    "request_key",
    "invalidate",
    "set_timeout",
]

import ctypes
import ctypes.util
import errno
import os
from array import array
from ctypes import (
    c_char_p,
    c_int32 as key_serial_t,
    c_long,
    c_size_t,
    c_uint,
    c_void_p,
)

KEY_SPEC_PROCESS_KEYRING = -2


class KeyutilsError(OSError):

    def __init__(self, code):
        self.code = code
        self.name = errno.errorcode.get(code, code)
        self.text = os.strerror(code)
        super().__init__(self.code, self.name, self.text)


class KeyNotAvailable(KeyutilsError):
    pass


class KeyExpired(KeyutilsError):
    pass


class KeyRevoked(KeyutilsError):
    pass


class KeyRejected(KeyutilsError):
    pass


_errno_map = {
    errno.ENOKEY: KeyNotAvailable,
    errno.EKEYEXPIRED: KeyExpired,
    errno.EKEYREVOKED: KeyRevoked,
    errno.EKEYREJECTED: KeyRejected,
}


def _errcheck(result):
    if result < 0:
        err_code = ctypes.get_errno()
        exc_type = _errno_map.get(err_code, KeyutilsError)
        raise exc_type(err_code)
    return result


def _declare(lib, name, restype, argtypes):
    try:
        func = lib[name]
    except (AttributeError, KeyError) as e:
        raise ImportError("Missing symbol in library 'keyutils.so'.") from e
    func.restype = restype
    func.argtypes = argtypes
    return func


_lib_name = ctypes.util.find_library("keyutils")
if _lib_name is None:
    raise ImportError("Library 'keyutils.so' not found")

_keyutils = ctypes.CDLL(_lib_name, use_errno=True)

_add_key = _declare(_keyutils, "add_key", key_serial_t, [
    c_char_p,       # [in] type
    c_char_p,       # [in] description
    c_void_p,       # [in] payload
    c_size_t,       # [in] plen
    key_serial_t,   # [in] ringid
])

_request_key = _declare(_keyutils, "request_key", key_serial_t, [
    c_char_p,       # [in] type
    c_char_p,       # [in] description
    c_char_p,       # [in] callout_info
    key_serial_t,   # [in] destringid
])

_keyctl_read = _declare(_keyutils, "keyctl_read", c_long, [
    key_serial_t,   # [in] id
    c_char_p,       # [out] buffer
    c_size_t,       # [in] buflen
])

_keyctl_invalidate = _declare(_keyutils, "keyctl_invalidate", c_long, [
    key_serial_t,   # [in] id
])

_keyctl_set_timeout = _declare(_keyutils, "keyctl_set_timeout", c_long, [
    key_serial_t,   # [in] id
    c_uint,         # [in] timeout
])

_keyctl_clear = _declare(_keyutils, "keyctl_clear", c_long, [
    key_serial_t,   # [in] ringid
])


def add_key(
    key: bytes,
    value: bytes,
    keyring: int = KEY_SPEC_PROCESS_KEYRING
) -> int:
    """
    Store secret with content.

    :param bytes key: key name
    :param bytes value: secret content
    :param int keyring: key ring ID
    :returns: ID of the inserted key
    :raises: KeyutilsError
    """
    return _errcheck(_add_key(b"user", key, value, len(value), keyring))


def is_valid(key: int) -> bool:
    """
    Check if the given key is available and valid.

    :param int key: key ID
    :returns: whether the key is accessible
    """
    return _keyctl_read(key, None, 0) >= 0


def read_key(key_id: int) -> bytes:
    """
    Read key content.

    :param int key_id: key ID
    :returns: secret content
    :raises: KeyutilsError
    """
    buflen = _errcheck(_keyctl_read(key_id, None, 0))
    while True:
        buffer = ctypes.create_string_buffer(buflen)
        ret = _errcheck(_keyctl_read(key_id, buffer, buflen))
        if 0 <= ret <= buflen:
            return buffer.value
        buflen = ret


def request_key(key: bytes, keyring: int = KEY_SPEC_PROCESS_KEYRING) -> int:
    """
    Find key ID by name.

    :param bytes key: key name
    :param int keyring: key ring ID
    :returns: key ID
    :raises: KeyutilsError
    """
    return _errcheck(_request_key(b"user", key, c_char_p(), keyring))


def invalidate(key: int):
    """
    Remove the specified key from the keyring.

    :param int key: key ID
    :raises: KeyutilsError
    """
    _errcheck(_keyctl_invalidate(key))


def set_timeout(key: int, timeout: int):
    """
    Set timeout in seconds for the specified key.

    :param int key: key ID
    :param int timeout: timeout in seconds
    :raises: KeyutilsError
    """
    _errcheck(_keyctl_set_timeout(key, timeout))


def clear(keyring: int = KEY_SPEC_PROCESS_KEYRING):
    """
    Clear all keys in the specified keyring.

    :param int keyring: keyring ID
    :raises: KeyutilsError
    """
    _errcheck(_keyctl_clear(keyring))


def list_keys(keyring: int = KEY_SPEC_PROCESS_KEYRING) -> [int]:
    """
    List key IDs in the specified keyring.

    :param int keyring: keyring ID
    :returns: list of key IDs
    :raises: keyutilsError
    """
    try:
        content = read_key(keyring)
    except KeyNotAvailable:
        return []
    return array('I', content).tolist()
